# -*- coding: utf-8 -*-
import datetime

import pytest


def test_setup_and_info(store):
    # GIVEN a store which is already setup
    assert len(store.engine.table_names()) > 0
    # THEN it should contain an Info entry with a default "created_at" date
    info_obj = store.info()
    assert isinstance(info_obj.created_at, datetime.datetime)


def test_track_update(store):
    # GIVEN a store which is empty apart from the initialized info entry
    assert store.info().updated_at is None
    # WHEN updating the last updated date
    store.track_update()
    # THEN it should update info entry with the current date
    assert isinstance(store.info().updated_at, datetime.datetime)


def test_add_user(store):
    # GIVEN an empty database
    assert store.User.query.first() is None
    name, email = "Paul T. Anderson", "paul.anderson@magnolia.com"
    # WHEN adding a new user
    new_user = store.add_user(name, email)
    # THEN it should be stored in the database
    assert store.User.query.filter_by(email=email).first() == new_user


def test_user(store):
    # GIVEN a database with a user
    name, email = "Paul T. Anderson", "paul.anderson@magnolia.com"
    store.add_user(name, email)
    assert store.User.query.filter_by(email=email).first().email == email
    # WHEN querying for a user
    user_obj = store.user(email)
    # THEN it should be returned
    assert user_obj.email == email

    # WHEN querying for a user that doesn't exist
    user_obj = store.user("this_is_a_made_up_email@fake_example.com")
    # THEN it should return as None
    assert user_obj is None


def test_analysis(sample_store):
    # GIVEN a store with an analysis
    existing_analysis = sample_store.analyses().first()
    # WHEN accessing it by ID
    analysis_obj = sample_store.analysis(existing_analysis.id)
    # THEN it should return the same analysis
    assert analysis_obj == existing_analysis

    # GIVEN an id that doesn't exist
    missing_analysis_id = 12312423534
    # WHEN accessing the analysis
    analysis_obj = sample_store.analysis(missing_analysis_id)
    # THEN it should return None
    assert analysis_obj is None


@pytest.mark.parametrize(
    "family, expected_bool",
    [
        ("crazygoat", True),  # running
        ("nicemouse", False),  # completed
        ("politesnake", False),  # failed
        ("gentlebird", True),  # pending
    ],
)
def test_is_latest_analysis_ongoing(sample_store, family, expected_bool):
    # GIVEN an analysis
    analysis_objs = sample_store.analyses(family=family).first()
    assert analysis_objs is not None
    # WHEN checking if the family has an ongoing analysis status
    is_ongoing = sample_store.is_latest_analysis_ongoing(family)
    # THEN it should return the expected result
    assert is_ongoing is expected_bool


@pytest.mark.parametrize(
    "family, expected_bool",
    [
        ("crazygoat", False),  # running
        ("nicemouse", False),  # completed
        ("politesnake", True),  # failed
        ("gentlebird", False),  # pending
    ],
)
def test_is_latest_analysis_failed(sample_store, family, expected_bool):
    # GIVEN an analysis
    analysis_objs = sample_store.analyses(family=family).first()
    assert analysis_objs is not None
    # WHEN checking if the family has a failed analysis status
    is_failed = sample_store.is_latest_analysis_failed(family)
    # THEN it should return the expected result
    assert is_failed is expected_bool


@pytest.mark.parametrize(
    "family, expected_bool",
    [
        ("crazygoat", False),  # running
        ("nicemouse", True),  # completed
        ("politesnake", False),  # failed
        ("gentlebird", False),  # pending
    ],
)
def test_is_latest_analysis_completed(sample_store, family, expected_bool):
    # GIVEN an analysis
    analysis_objs = sample_store.analyses(family=family).first()
    assert analysis_objs is not None
    # WHEN checking if the family has a failed analysis status
    is_failed = sample_store.is_latest_analysis_completed(family)
    # THEN it should return the expected result
    assert is_failed is expected_bool


@pytest.mark.parametrize(
    "family, expected_status",
    [
        ("crazygoat", "running"),
        ("nicemouse", "completed"),
        ("politesnake", "failed"),
        ("gentlebird", "pending"),
    ],
)
def test_get_latest_analysis_status(sample_store, family, expected_status):
    # GIVEN an analysis
    analysis_objs = sample_store.analyses(family=family).first()
    assert analysis_objs is not None
    # WHEN checking if the family has an analysis status
    status = sample_store.get_latest_analysis_status(family)
    # THEN it should return the expected result
    assert status is expected_status


def test_aggregate_jobs(sample_store):

    # GIVEN a store with some analyses
    assert sample_store.analyses().count() > 0
    all_jobs = sample_store.jobs().count()
    assert all_jobs == 2

    # WHEN aggregating data on failed jobs
    jobs_data = sample_store.aggregate_failed()

    # THEN it should return a list of dicts per job type with count
    assert isinstance(jobs_data, list)

    # ... it should exclude "cancelled" jobs
    assert len(jobs_data) == 1
    assert jobs_data[0]["name"] == "samtools_mpileup"
    assert jobs_data[0]["count"] == 1


def test_aggregate_jobs_since_forever_date(sample_store):

    # GIVEN a store with some analyses
    assert sample_store.analyses().count() > 0
    all_jobs = sample_store.jobs().count()
    assert all_jobs == 2
    # a date a gazillion days back
    date_since = datetime.datetime.now() - datetime.timedelta(days=9999)

    # WHEN aggregating data on failed jobs
    jobs_data = sample_store.aggregate_failed(date_since)

    # THEN it should return a list of dicts per job type with count
    assert len(jobs_data) == 1
    assert jobs_data[0]["name"] == "samtools_mpileup"
    assert jobs_data[0]["count"] == 1


def test_aggregate_jobs_since_yesterday(sample_store):

    # GIVEN a store with some analyses
    assert sample_store.analyses().count() > 0
    all_jobs = sample_store.jobs().count()
    assert all_jobs == 2
    # a date a gazillion days back
    date_since = datetime.datetime.now() - datetime.timedelta(days=1)

    # WHEN aggregating data on failed jobs
    jobs_data = sample_store.aggregate_failed(date_since)

    # THEN it should return a list of dicts per job type with count
    assert len(jobs_data) == 0



# -*- coding: utf-8 -*-
import pytest

from functools import partial

from click.testing import CliRunner
import ruamel.yaml

from trailblazer.cli import base
from trailblazer.mip import sacct, files as files_api
from trailblazer.log import LogAnalysis
from trailblazer.store import Store


@pytest.fixture
def cli_runner():
    runner = CliRunner()
    return runner


@pytest.fixture
def invoke_cli(cli_runner):
    return partial(cli_runner.invoke, base)


@pytest.fixture(scope="session")
def files():
    return {
        "config": "tests/fixtures/case/case_config.yaml",
        "sampleinfo": "tests/fixtures/case/case_qc_sample_info.yaml",
        "qcmetrics": "tests/fixtures/case/case_qc_metrics.yaml",
        "sacct": "tests/fixtures/case/mip_2019-07-04T10:47:15.log.status",
    }


@pytest.fixture(scope="session")
def files_raw(files):
    return {
        "config": ruamel.yaml.safe_load(open(files["config"])),
        "sampleinfo": ruamel.yaml.safe_load(open(files["sampleinfo"])),
        "qcmetrics": ruamel.yaml.safe_load(open(files["qcmetrics"])),
        "sacct": list(open(files["sacct"])),
    }


@pytest.fixture(scope="session")
def files_data(files_raw):
    return {
        "config": files_api.parse_config(files_raw["config"]),
        "sampleinfo": files_api.parse_sampleinfo(files_raw["sampleinfo"]),
        "qcmetrics": files_api.parse_qcmetrics(files_raw["qcmetrics"]),
        "sacct": sacct.parse_sacct(files_raw["sacct"]),
    }


@pytest.fixture(scope="session")
def allfailed_sacct_jobs():
    with open("tests/fixtures/sacct/allfailed.log.status") as stream:
        sacct_jobs = sacct.parse_sacct(stream)
    return sacct_jobs


@pytest.yield_fixture(scope="function")
def store(tmpdir):
    _store = Store(uri="sqlite://", families_dir=tmpdir)
    _store.setup()
    yield _store
    _store.drop_all()


@pytest.yield_fixture(scope="function")
def sample_store(store):
    sample_data = ruamel.yaml.safe_load(open("tests/fixtures/sample-data.yaml"))
    for user_data in sample_data["users"]:
        store.add_user(user_data["name"], user_data["email"])
    for analysis_data in sample_data["analyses"]:
        analysis_data["user"] = store.user(analysis_data["user"])
        failed_jobs = analysis_data.get("failed_jobs", [])
        analysis_data["failed_jobs"] = [store.Job(**job_data) for job_data in failed_jobs]
        store.add(store.Analysis(**analysis_data))
    store.commit()
    store.add_pending(family="gentlebird", email="tom.cruise@magnolia.com")
    yield store


@pytest.fixture
def log_analysis(store):
    _log_analysis = LogAnalysis(store)
    return _log_analysis



# -*- coding: utf-8 -*-
import os

import coloredlogs
from flask import Flask
from flask_cors import CORS
from flask_reverse_proxy import FlaskReverseProxied

from trailblazer.server import api, ext

coloredlogs.install(level="INFO")
app = Flask(__name__)

SECRET_KEY = "unsafe!!!"
TEMPLATES_AUTO_RELOAD = True
SQLALCHEMY_DATABASE_URI = os.environ["SQLALCHEMY_DATABASE_URI"]
SQLALCHEMY_POOL_RECYCLE = 7200
SQLALCHEMY_TRACK_MODIFICATIONS = "FLASK_DEBUG" in os.environ

app.config.from_object(__name__)

# register blueprints
app.register_blueprint(api.blueprint)

# configure extensions
FlaskReverseProxied(app)
ext.store.init_app(app)
cors = CORS(app, resources={r"/api/*": {"origins": "*"}}, supports_credentials=True)


@app.route("/")
def index():
    return "Welcome to Trailblazer REST API"



"""Test all aspects of the get command"""


def test_get(invoke_cli):
    """Test get with a argument that should yield no results"""
    # GIVEN the cli
    # WHEN running get with a comment
    result = invoke_cli(["get", "-c", "dummy-comment"])
    # THEN all analyses containing that comment should be listed
    assert not result.output



def job_ids(log_stream):
    """Grep out all lines with scancel example."""
    id_rows = [line for line in log_stream if "scancel" in line]
    jobs = [id_row.strip()[-7:-1] for id_row in id_rows]
    return jobs



# -*- coding: utf-8 -*-
import datetime

import pytest

from trailblazer.mip import sacct


@pytest.mark.parametrize(
    "time_str, expected_seconds",
    [
        ("02:53:08", (2 * 60 * 60) + (53 * 60)),
        ("3-07:13:56", (3 * 24 * 60 * 60) + (7 * 60 * 60) + (13 * 60)),
        ("18:47.020", (18 * 60)),
        ("00:00:35", 0),
    ],
)
def test_time_to_sec(time_str, expected_seconds):
    # GIVEN a time string
    # WHEN parsing it into number of seconds
    seconds = sacct.time_to_sec(time_str)
    # THEN it should match expected number of seconds (skipping individual secs)
    assert seconds == expected_seconds


@pytest.mark.parametrize(
    "row, expected_job",
    [
        (
            [
                "813887",
                "gatk_genotypegvcfs_case_BOTH",
                "prod001",
                "core",
                "16",
                "10:03.096",
                "00:02:01",
                "2017-06-19T00:44:46",
                "2017-06-19T00:46:47",
                "COMPLETED",
                "0:0",
            ],
            dict(step="gatk_genotypegvcfs", context="case", is_completed=True),
        ),
        (
            [
                "815038",
                "fastqc_ACC2698A3",
                "prod001",
                "core",
                "8",
                "00:00:00",
                "00:37:24",
                "2017-06-20T12:57:45",
                "Unknown",
                "RUNNING",
                "0:0",
            ],
            dict(step="fastqc", context="ACC2698A3", is_completed=False),
        ),
    ],
)
def test_convert_job(row, expected_job):
    # GIVEN a sacct row
    # WHEN parsing it for extracted data
    job = sacct.convert_job(row)
    # THEN it should have an integer for a job id
    assert isinstance(job["id"], int)
    # ... the step should be the job type
    assert job["step"] == expected_job["step"]
    # ... the context should be the family/sample id
    assert job["context"] == expected_job["context"]
    # ... it should be marked as completed only if that is the state
    assert job["is_completed"] == expected_job["is_completed"]
    # ... the "start" should be a date
    assert isinstance(job["start"], datetime.datetime)
    # ... the "end" should be a date unless the job is "running"
    if job["state"] == "RUNNING":
        assert job["end"] is None
    else:
        assert isinstance(job["end"], datetime.datetime)


def test_parse_sacct(files_raw):
    # GIVEN a sacct file with a number of row (1 header, rest jobs)
    lines = len(files_raw["sacct"])
    # WHEN parsing the sacct file
    jobs = sacct.parse_sacct(files_raw["sacct"])
    # THEN it should find all relevant rows
    assert len(jobs) == (lines - 1)


def test_filter_jobs(failed_sacct_jobs):
    # GIVEN a sacct output with COMPLETED, FAILED, and CANCELLED jobs
    completed_count = 17
    failed_count = 2
    cancelled_count = 34
    # WHEN filtering on "failed" jobs
    failed_jobs = sacct.filter_jobs(failed_sacct_jobs, failed=True)
    # THEN it should return jobs which are "failed" or "cancelled"
    assert len(failed_jobs) == (failed_count + cancelled_count)

    # WHEN filtering on "OK" jobs
    ok_jobs = sacct.filter_jobs(failed_sacct_jobs, failed=False)
    # THEN it should return jobs which are "completed"
    assert len(ok_jobs) == completed_count



# -*- coding: utf-8 -*-
import logging
from pathlib import Path
from typing import List

import ruamel.yaml

from trailblazer.mip import sacct as sacct_api, files as files_api, miplog
from trailblazer.store import models, Store
from trailblazer.exc import MissingFileError

log = logging.getLogger(__name__)


class LogAnalysis(object):
    def __init__(self, store: Store):
        self.store = store

    def __call__(self, config_stream: List[str], sampleinfo: str = None, sacct: str = None):
        """Add a new analysis log."""
        config_raw = ruamel.yaml.safe_load(config_stream)
        config_data = files_api.parse_config(config_raw)
        sampleinfo_path = Path(sampleinfo or config_data["sampleinfo_path"])
        if not sampleinfo_path.exists():
            raise MissingFileError(sampleinfo_path)
        with sampleinfo_path.open() as stream:
            sampleinfo_raw = ruamel.yaml.safe_load(stream)
        sampleinfo_data = files_api.parse_sampleinfo_light(sampleinfo_raw)
        sacct_path = Path(sacct if sacct else f"{config_data['log_path']}.status")
        if not sacct_path.exists():
            raise MissingFileError(sacct_path)
        with sacct_path.open() as stream:
            sacct_jobs = sacct_api.parse_sacct(stream)
        with Path(config_data["log_path"]).open() as stream:
            jobs = len(miplog.job_ids(stream))
        run_data = self.parse(config_data, sampleinfo_data, sacct_jobs, jobs=jobs)
        self._delete_temp_logs(run_data["case"])
        new_run = self.build(run_data)
        if new_run:
            self.store.add_commit(new_run)
            return new_run

    def _delete_temp_logs(self, family_name: str):
        """Delete temporary logs for the current family."""
        for temp_log in self.store.analyses(family=family_name, temp=True):
            log.debug(f"delete temporary log: {temp_log.id} - {temp_log.status}")
            temp_log.delete()

    @classmethod
    def parse(
        cls, config_data: dict, sampleinfo_data: dict, sacct_jobs: List[dict], jobs: int
    ) -> dict:
        """Parse information about a run."""
        analysis_types = [sample["type"] for sample in config_data["samples"]]
        run_data = {
            "user": config_data["email"],
            "case": config_data["case"],
            "priority": config_data["priority"],
            "started_at": sampleinfo_data["date"],
            "version": sampleinfo_data["version"],
            "out_dir": config_data["out_dir"],
            "config_path": config_data["config_path"],
            "type": cls._get_analysis_type(analysis_types),
        }

        sacct_data, last_job_end = cls._parse_sacct(sacct_jobs, jobs_count=jobs)
        run_data.update(sacct_data)

        run_data["status"] = cls.get_status(
            sampleinfo_data["is_finished"], len(run_data["failed_jobs"])
        )
        if run_data["status"] == "completed":
            run_data["completed_at"] = last_job_end

        return run_data

    @staticmethod
    def _parse_sacct(sacct_jobs: List[dict], jobs_count: int = None):
        """Parse out info from Sacct log."""
        failed_jobs = sacct_api.filter_jobs(sacct_jobs, failed=True)
        completed_jobs = [job for job in sacct_jobs if job["is_completed"]]
        last_job_end = completed_jobs[-1]["end"] if len(completed_jobs) > 0 else None
        data = {
            "jobs": jobs_count,
            "completed_jobs": len(completed_jobs),
            "progress": (len(completed_jobs) / jobs_count) if jobs_count else None,
            "failed_jobs": [
                {
                    "slurm_id": job["id"],
                    "started_at": job["start"],
                    "elapsed": job["elapsed"],
                    "status": job["state"].lower(),
                    "name": job["step"],
                    "context": job["context"],
                }
                for job in failed_jobs
            ],
        }
        return data, last_job_end

    @staticmethod
    def get_status(finished: str, failed_jobs: int) -> str:
        """Determine status of a run."""
        if finished:
            return "completed"
        elif failed_jobs > 0:
            return "failed"
        else:
            return "running"

    @staticmethod
    def _get_analysis_type(analysis_types: List[str]) -> str:
        """Determine the overall analysis type."""
        types_set = set(analysis_types)
        return types_set.pop() if len(types_set) == 1 else "wgs"

    def build(self, run_data: dict) -> models.Analysis:
        """Build a new Analysis object."""
        existing_run = self.store.find_analysis(
            family=run_data["case"],
            started_at=run_data["started_at"],
            status=run_data["status"],
        )
        if existing_run:
            return None

        run_data["user"] = self.store.user(run_data["user"])
        new_failed_jobs = [self.store.Job(**job) for job in run_data["failed_jobs"]]
        del run_data["failed_jobs"]
        new_run = self.store.Analysis(family=run_data["case"], **run_data)
        new_run.failed_jobs = new_failed_jobs
        return new_run



# -*- coding: utf-8 -*-
from trailblazer.store import models


def test_user_first_name():
    # GIVEN a user with a name
    user_obj = models.User(name="Paul T. Anderson")
    # WHEN accessing the first name
    first_name = user_obj.first_name
    # THEN it should return the spoken name
    assert first_name == "Paul"



# -*- coding: utf-8 -*-
import datetime
from pathlib import Path

import pytest

from trailblazer import log, exc
from trailblazer.log import LogAnalysis


@pytest.mark.parametrize(
    "kwargs, expected_status",
    [
        (dict(finished=False, failed_jobs=0), "running"),
        (dict(finished=True, failed_jobs=0), "completed"),
        (dict(finished=False, failed_jobs=3), "failed"),
        (dict(finished=True, failed_jobs=3), "completed"),  # allows override!
    ],
)
def test_get_status(kwargs, expected_status):
    # GIVEN the inputs of wheter the run is finished and how many jobs have failed
    # WHEN getting the run status...
    run_status = log.LogAnalysis.get_status(**kwargs)
    # THEN it should return the expected status
    assert run_status == expected_status


def test_call(store, log_analysis, files):

    # GIVEN an empty store
    assert store.analyses().count() == 0

    # WHEN adding a new analysis log entry
    with Path(files["config"]).open() as config_stream:
        new_run = log_analysis(config_stream, sampleinfo=files["sampleinfo"], sacct=files["sacct"])

    # THEN it should add the analysis log
    assert store.analyses(family=new_run.family).first() == new_run


def test_call_twice(store, log_analysis, files):
    # GIVEN an empty store
    assert store.analyses().count() == 0

    # WHEN adding a new analysis log entry twice
    with Path(files["config"]).open() as config_stream:
        log_analysis(config_stream, sampleinfo=files["sampleinfo"], sacct=files["sacct"])
    first_analysis = store.analyses().first()
    with Path(files["config"]).open() as config_stream:
        new_run = log_analysis(config_stream, sampleinfo=files["sampleinfo"], sacct=files["sacct"])

    # THEN it should skip adding the duplicate log to the database
    assert new_run is None
    assert store.analyses(family=first_analysis.family).count() == 1
    assert store.analyses(family=first_analysis.family).first() == first_analysis


def test_call_with_missing_files(log_analysis, files):
    # GIVEN missing sampleinfo and existing sacct
    config_path = Path(files["config"])
    sampleinfo_path = Path("/tmp/I_DO_NOT_EXIST.txt")
    sacct_path = Path(files["sacct"])
    assert sampleinfo_path.exists() is False
    assert sacct_path.exists() is True
    # WHEN calling the log analysis API
    # THEN it should raise an exception
    with pytest.raises(exc.MissingFileError):
        with config_path.open() as config_stream:
            log_analysis(config_stream, sampleinfo=sampleinfo_path, sacct=sacct_path)

    # GIVEN existing sampleinfo and missing sacct
    sampleinfo_path = Path(files["sampleinfo"])
    sacct_path = Path("/tmp/I_DO_NOT_EXIST.txt")
    assert sampleinfo_path.exists() is True
    assert sacct_path.exists() is False
    # WHEN calling the log analysis API
    # THEN it should also raise and exception
    with pytest.raises(exc.MissingFileError):
        with config_path.open() as config_stream:
            log_analysis(config_stream, sampleinfo=sampleinfo_path, sacct=sacct_path)


def test_parse_sacct(files_data):
    # GIVEN sacct jobs from a finished analysis
    sacct_jobs = files_data["sacct"]
    analysis_end = datetime.datetime(2019, 7, 4, 23, 16, 54)
    # WHEN parsing them for info
    jobs_count = len(sacct_jobs)  # TODO: see code in _call_ in class LogAnalysis for proper
    # jobs_count calculation
    sacct_data, last_job_end = log.LogAnalysis._parse_sacct(sacct_jobs, jobs_count=jobs_count)

    # THEN it should return info about the jobs in general
    assert sacct_data["jobs"] == len(sacct_jobs)
    assert sacct_data["completed_jobs"] + len(sacct_data["failed_jobs"]) == len(sacct_jobs)

    # ... and correctly determine the end of the last job
    assert last_job_end == analysis_end


def test_parse_sacct_all_fails(allfailed_sacct_jobs):
    # GIVEN sacct jobs from analysis with all failed jobs
    # WHEN parsing out info
    sacct_data, last_job_end = log.LogAnalysis._parse_sacct(allfailed_sacct_jobs)
    # THEN the "analysis end" time should be unknown
    assert last_job_end is None


@pytest.mark.parametrize(
    "analysis_types, expected_type",
    [
        (["wes"], "wes"),  # single sample and analysis type
        (["wgs", "wgs", "wgs"], "wgs"),  # multiple samples, single type
        (["wes", "wgs", "wgs"], "wgs"),  # multiple samples, multiple types
    ],
)
def test_get_analysis_type(analysis_types, expected_type):
    # GIVEN a list of analysis types; wes, wgs, rna
    # WHEN determining the overall type
    analysis_type = log.LogAnalysis._get_analysis_type(analysis_types)
    # THEN it should be correct :-P
    assert analysis_type == expected_type


def test_parse(files_data):
    # GIVEN a finished analysis with all jobs completed
    config_data = files_data["config"]
    sampleinfo_data = files_data["sampleinfo"]
    sacct_jobs = files_data["sacct"]
    # WHEN parsing log-information
    run_data = log.LogAnalysis.parse(config_data, sampleinfo_data, sacct_jobs, jobs=None)
    # THEN it should have parsed out a completed date
    assert run_data["status"] == "completed"
    assert isinstance(run_data["completed_at"], datetime.datetime)



import click


@click.command("get")
@click.option("-c", "--comment", help="return analyses with comments containing")
@click.pass_context
def get(context, comment):
    """Display recent logs for analyses."""
    runs = context.obj["store"].find_analyses_with_comment(comment=comment).limit(100)
    for run_obj in runs:
        if run_obj.status == "pending":
            message = f"{run_obj.id} | {run_obj.family} [{run_obj.status.upper()}]"
        else:
            message = (
                f"{run_obj.id} | {run_obj.family} {run_obj.started_at.date()} "
                f"[{run_obj.type.upper()}/{run_obj.status.upper()}]"
            )
            if run_obj.status == "running":
                message = click.style(f"{message} - {run_obj.progress * 100}/100", fg="blue")
            elif run_obj.status == "completed":
                message = click.style(f"{message} - {run_obj.completed_at}", fg="green")
            elif run_obj.status == "failed":
                message = click.style(message, fg="red")

            message += f" {run_obj.comment}"
        click.echo(message)



"""Test MIP files"""

import copy
import dateutil
import pytest

from trailblazer.mip import files


@pytest.mark.parametrize(
    "config, expected_case",
    [
        ({"case_id": "crazygoat"}, "crazygoat"),
        ({"family_id": "crazygoat"}, "crazygoat"),
        ({"not_a_case_id": "anonymous"}, None),
    ],
)
def test_get_case_from_config(config: dict, expected_case: str):
    """Test getting the case id from mip config"""
    # GIVEN a case id or family id

    # WHEN getting case from config dict
    case = files.get_case_from_config(config=config)

    # THEN return case
    assert case == expected_case


@pytest.mark.parametrize(
    "config, expected_bool",
    [
        ({"dry_run_all": 1}, True),
        ({"not_a_dry_run": 1}, False),
    ],
)
def test_get_dry_run_all(config: dict, expected_bool: bool):
    """Test getting the dry_run_all from mip config"""
    # GIVEN a dry run mip analysis

    # WHEN getting dry_run_all from config dict
    is_dry_run = files.get_dry_run_all(config=config)

    # THEN return true
    assert is_dry_run == expected_bool


def test_get_sample_data_from_config():
    """Test getting sample data from mip config"""
    # GIVEN sample data from a mip analysis config
    config = {
        "analysis_type": {"sample_1": "wgs"},
    }

    # WHEN getting sample data from config dict
    sample_data = files.get_sample_data_from_config(config=config)

    expected_sample_data = [
        {
            "id": "sample_1",
            "type": "wgs",
        }
    ]
    # THEN return true
    assert sample_data == expected_sample_data


def test_parse_config(files_raw: dict) -> dict:
    """
    Args:
    files_raw (dict): With dicts from files
    """

    # GIVEN config data of a "sharp" run (not dry run)
    config_raw = files_raw["config"]

    # WHEN parsing the MIP output config
    config_data = files.parse_config(config_raw)

    # THEN it return a dict
    assert isinstance(config_data, dict)

    # THEN the dry run is True
    assert config_data["is_dryrun"] is True

    # THEN the number of samples and analysis_types should be equal
    assert len(config_raw["analysis_type"]) == len(config_data["samples"])

    expected_config_data = {
        "case": "case",
        "config_path": "/path_to/cases/case/analysis/case_config.yaml",
        "is_dryrun": True,
        "email": "test.person@test.se",
        "log_path": "tests/fixtures/case/mip_2019-07-04T10:47:15.log",
        "out_dir": "/path_to/cases/case/analysis",
        "priority": "normal",
        "sampleinfo_path": "/path_to/cases/case/analysis/case_qc_sample_info.yaml",
    }

    # Then config data should be set
    for key, value in expected_config_data.items():
        assert config_data[key] == value


@pytest.mark.parametrize(
    "sample_info, expected_bool",
    [
        (
            {
                "analysisrunstatus": "finished",
            },
            True,
        ),
        (
            {
                "analysisrunstatus": "notfinished",
            },
            False,
        ),
    ],
)
def test_get_analysisrunstatus(sample_info: dict, expected_bool: bool):
    """Test getting analysis run status from mip config"""
    # GIVEN a run status from a sample_info file

    # WHEN getting sample data from config dict
    is_finished = files.get_analysisrunstatus(sample_info=sample_info)

    # THEN return true
    assert is_finished == expected_bool


def test_get_sampleinfo_date():
    """Test getting the case id from mip config"""
    # GIVEN a date
    sample_info = {
        "analysis_date": "1999-12-31",
    }

    # WHEN getting date from sample info
    date = files.get_sampleinfo_date(data=sample_info)

    # THEN return analysis date
    assert date == "1999-12-31"


def test_parse_sampleinfo_light(files_raw: dict):
    """
    Args:
    files_raw (dict): With dicts from files
    """

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = files_raw["sampleinfo"]

    # WHEN parsing the MIP output sampleinfo
    sampleinfo_data = files.parse_sampleinfo_light(sampleinfo_raw)

    # THEN it should return a dict
    assert isinstance(sampleinfo_data, dict)

    # THEN is_finished should be true
    assert sampleinfo_data["is_finished"] is True

    # THEN date should be set
    assert sampleinfo_data["date"] == dateutil.parser.parse("2019-07-05T10:12:03")

    # THEN version should be set
    assert sampleinfo_data["version"] == "v7.1.0"


def test_get_rank_model_version(files_raw: dict, mip_meta_data: dict):
    """Test getting the rank model from sample_info file"""

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = files_raw["sampleinfo"]

    # WHEN getting the rank_model
    rank_model_version = files.get_rank_model_version(
        sample_info=sampleinfo_raw, rank_model_type="rank_model", step="genmod"
    )

    # THEN the rank model version should be returned
    assert rank_model_version == mip_meta_data["RANK_MODEL_VERSION"]


def test_get_rank_model_version_with_program(files_raw: dict, mip_meta_data: dict):
    """Test getting the rank model from sample_info file with program key"""

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = copy.copy(files_raw["sampleinfo"])

    # Using old MIP style with program instead of recipe
    sampleinfo_raw["program"] = sampleinfo_raw.pop("recipe")

    # WHEN getting the rank_model
    rank_model_version = files.get_rank_model_version(
        sample_info=sampleinfo_raw, rank_model_type="rank_model", step="genmod"
    )

    # THEN the rank model version should be returned
    assert rank_model_version == mip_meta_data["RANK_MODEL_VERSION"]


@pytest.mark.parametrize(
    "sample_info, expected_case",
    [
        ({"case": "crazygoat"}, "crazygoat"),
        ({"family": "crazygoat"}, "crazygoat"),
        ({"not_a_case": "anonymous"}, None),
    ],
)
def test_get_case_from_sampleinfo(sample_info: dict, expected_case: str):
    """Test getting the case or family from mip sampleinfo file"""
    # GIVEN a case or family

    # WHEN getting case from config dict
    case = files.get_case_from_sampleinfo(sample_info=sample_info)

    # THEN return case
    assert case == expected_case


def test_get_genome_build(mip_meta_data: dict):
    """Test getting sample data from mip config"""
    # GIVEN human genome build in sample info
    sample_info = {
        "human_genome_build": {
            "source": mip_meta_data["GENOME_BUILD_SOURCE"],
            "version": mip_meta_data["GENOME_BUILD_VERSION"],
        },
    }

    # WHEN getting sample data from config dict
    genome_build = files.get_genome_build(sample_info=sample_info)

    # THEN return genome build (source and version)
    assert genome_build == "grch37"


def test_parse_sampleinfo(files_raw: dict, mip_meta_data: dict):
    """
    Args:
    files_raw (dict): With dicts from files
    """

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = files_raw["sampleinfo"]

    # WHEN parsing the MIP output sampleinfo
    sampleinfo_data = files.parse_sampleinfo(sampleinfo_raw)

    # THEN it should return a dict
    assert isinstance(sampleinfo_data, dict)

    # THEN is_finished should be True
    assert sampleinfo_data["is_finished"] is True

    expected_sampleinfo_data = {
        "case": "case",
        "genome_build": "grch37",
        "version": "v7.1.0",
    }

    # Then sampleinfo case data should be set
    for key, value in expected_sampleinfo_data.items():
        assert sampleinfo_data[key] == value

    expected_sampleinfo_sample_data = {"id": "mother"}

    # Then sampleinfo sample data should be set
    for key, value in expected_sampleinfo_sample_data.items():
        for sample_data in sampleinfo_data["samples"]:

            if sample_data["id"] != expected_sampleinfo_sample_data["id"]:
                continue

            assert sample_data[key] == value

    # THEN the number of samples and analysis_types should be equal
    assert len(sampleinfo_raw["analysis_type"]) == len(sampleinfo_data["samples"])


def test_set_bamstats_metrics_single_bamstat():
    """Test setting bam stats from qc_metrics when single bamstat entry"""
    # GIVEN 2 bamstats file metrics for a sample
    file_metric = {"bamstats": {"raw_total_sequences": 2, "reads_mapped": 1}}
    sample_data = {}

    # WHEN setting bamstats metrics
    sample_data = files.set_bamstats_metrics(file_metrics=file_metric, sample_data=sample_data)

    expected_sample_data = {"reads": 2, "total_mapped": 1}
    # THEN return bam stats metrics in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_bamstats_metrics_mutiple_bamstat():
    """Test setting bam stats from qc_metrics when multiple bamstat entry"""
    # GIVEN 2 bamstats file metrics for a sample
    sample_metric = {
        "file_1": {"bamstats": {"raw_total_sequences": 2, "reads_mapped": 1}},
        "file_2": {"bamstats": {"raw_total_sequences": 1, "reads_mapped": 2}},
    }
    sample_data = {}

    for file_metric in sample_metric:
        # WHEN setting bamstats metrics
        sample_data = files.set_bamstats_metrics(
            file_metrics=sample_metric[file_metric], sample_data=sample_data
        )

    expected_sample_data = {"reads": 3, "total_mapped": 3}
    # THEN return bam stats metrics in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_chanjo_sexcheck_metrics(files_raw: dict):
    """Test setting gender from file metric for chanjo_sexcheck"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. markduplicates
    sample_metrics = qcmetrics_raw["sample"]["father"]["father_lanes_1_sorted_md_brecal_sex"]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_chanjo_sexcheck_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {"predicted_sex": "male"}
    # THEN return predicted sex in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_collecthsmetrics_metrics(files_raw: dict, mip_meta_data: dict):
    """Test setting hsmetrics metrics from file metric for collecthsmetrics"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. markduplicates
    sample_metrics = qcmetrics_raw["sample"]["father"][
        "father_lanes_1_sorted_md_brecal_collecthsmetrics"
    ]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_collecthsmetrics_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {
        "at_dropout": mip_meta_data["FATHER_AT_DROPOUT"],
        "completeness_target": {
            10: mip_meta_data["FATHER_PCT_TARGET_BASES_10X"],
            20: mip_meta_data["FATHER_PCT_TARGET_BASES_20X"],
            50: mip_meta_data["FATHER_PCT_TARGET_BASES_50X"],
            100: mip_meta_data["FATHER_PCT_TARGET_BASES_100X"],
        },
        "gc_dropout": mip_meta_data["FATHER_GC_DROPOUT"],
        "target_coverage": mip_meta_data["FATHER_MEAN_TARGET_COVERAGE"],
    }
    # THEN return hs metrics in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_collectmultiplemetrics_metrics(files_raw: dict, mip_meta_data: dict):
    """Test setting multiple metrics from file metric for collectmultiplemetrics"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. collectmultiplemetrics
    sample_metrics = qcmetrics_raw["sample"]["father"][
        "father_lanes_1_sorted_md_brecal_collectmultiplemetrics"
    ]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_collectmultiplemetrics_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {"strand_balance": mip_meta_data["FATHER_STRAND_BALANCE"]}
    # THEN return strand balance in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_collectmultiplemetricsinsertsize_metrics(files_raw: dict, mip_meta_data: dict):
    """Test setting multiple metrics from file metric for collectmultiplemetricsinsertsize"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. collectmultiplemetricsinsertsize
    sample_metrics = qcmetrics_raw["sample"]["father"][
        "father_lanes_1_sorted_md_brecal_collectmultiplemetrics"
    ]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_collectmultiplemetricsinsertsize_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {
        "median_insert_size": mip_meta_data["FATHER_MEDIAN_INSERT_SIZE"],
        "insert_size_standard_deviation": mip_meta_data["FATHER_STANDARD_DEVIATION"],
    }
    # THEN return insert size metrics in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_markduplicates_metrics(files_raw: dict, mip_meta_data: dict):
    """Test setting duplicates metrics from file metric for markduplicates"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. markduplicates
    sample_metrics = qcmetrics_raw["sample"]["father"]["father_lanes_1_sorted_md"]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_markduplicates_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {"duplicates": mip_meta_data["FATHER_FRACTION_DUPLICATES"]}
    # THEN return duplicates in sample data for file metric
    assert sample_data == expected_sample_data


def test_get_sample_metrics(files_raw: dict, mip_meta_data: dict):
    """Test getting sample data for duplicates from file metrics for markduplicates"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. markduplicates
    sample_metrics = {
        "father_lanes_1_sorted_md": qcmetrics_raw["sample"]["father"]["father_lanes_1_sorted_md"]
    }
    sample_data = {}

    # WHEN getting sample metric for file
    sample_data = files.get_sample_metrics(sample_metrics=sample_metrics, sample_data=sample_data)

    expected_sample_data = {"duplicates": mip_meta_data["FATHER_FRACTION_DUPLICATES"]}
    # THEN return duplicates in sample data for file metric
    assert sample_data == expected_sample_data


def test_parse_qcmetrics(files_raw: dict, mip_meta_data: dict):
    """
    Args:
    files_raw (dict): With dicts from files
    """

    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # WHEN parsing qc metric data
    qcmetrics_data = files.parse_qcmetrics(qcmetrics_raw)

    # THEN it should return a dict
    assert isinstance(qcmetrics_data, dict)

    # Sample data
    # Build dict for sample return data
    expected_qcmetrics_sample_data = {
        "at_dropout": mip_meta_data["MOTHER_AT_DROPOUT"],
        "duplicates": mip_meta_data["MOTHER_FRACTION_DUPLICATES"],
        "id": "mother",
        "insert_size_standard_deviation": mip_meta_data["MOTHER_STANDARD_DEVIATION"],
        "gc_dropout": mip_meta_data["MOTHER_GC_DROPOUT"],
        "mapped": mip_meta_data["MOTHER_MAPPED"],
        "median_insert_size": mip_meta_data["MOTHER_MEDIAN_INSERT_SIZE"],
        "predicted_sex": "female",
        "reads": mip_meta_data["MOTHER_RAW_TOTAL_SEQUENCES"],
        "strand_balance": mip_meta_data["MOTHER_STRAND_BALANCE"],
        "target_coverage": mip_meta_data["MOTHER_MEAN_TARGET_COVERAGE"],
    }

    # THEN qcmetrics sample metrics should be set
    for key, value in expected_qcmetrics_sample_data.items():
        for sample_data in qcmetrics_data["samples"]:

            if sample_data["id"] != expected_qcmetrics_sample_data["id"]:
                continue

            assert sample_data[key] == value

    expected_qcmetrics_sample_cov_data = {
        10: mip_meta_data["MOTHER_PCT_TARGET_BASES_10X"],
        20: mip_meta_data["MOTHER_PCT_TARGET_BASES_20X"],
        50: mip_meta_data["MOTHER_PCT_TARGET_BASES_50X"],
        100: mip_meta_data["MOTHER_PCT_TARGET_BASES_100X"],
    }

    # THEN qcmetrics sample coverage metrics should be set
    for key, value in expected_qcmetrics_sample_cov_data.items():
        for sample_data in qcmetrics_data["samples"]:

            if sample_data["id"] != expected_qcmetrics_sample_data["id"]:
                continue

            assert sample_data["completeness_target"][key] == value



# -*- coding: utf-8 -*-
"""Store backend in Trailblazer"""
from typing import List
import datetime as dt
from pathlib import Path
import shutil

import alchy
import sqlalchemy as sqa
import ruamel

from trailblazer.constants import COMPLETED_STATUS, FAILED_STATUS, ONGOING_STATUSES
from trailblazer.store import models
from trailblazer.mip import files, trending


class BaseHandler:
    User = models.User
    Analysis = models.Analysis
    Job = models.Job
    Info = models.Info

    def setup(self):
        self.create_all()
        # add initial metadata record (for web interface)
        new_info = self.Info()
        self.add_commit(new_info)

    def find_analysis(self, family, started_at, status):
        """Find a single analysis."""
        query = self.Analysis.query.filter_by(family=family, started_at=started_at, status=status)
        return query.first()

    def find_analyses_with_comment(self, comment):
        """Find a analyses containing comment."""
        analysis_query = self.Analysis.query

        analysis_query = analysis_query.filter(self.Analysis.comment.like(f"%{comment}%"))
        return analysis_query

    def analyses(
        self,
        *,
        family: str = None,
        query: str = None,
        status: str = None,
        deleted: bool = None,
        temp: bool = False,
        before: dt.datetime = None,
        is_visible: bool = None,
    ):
        """Fetch analyses form the database."""
        analysis_query = self.Analysis.query
        if family:
            analysis_query = analysis_query.filter_by(family=family)
        elif query:
            analysis_query = analysis_query.filter(
                sqa.or_(
                    self.Analysis.family.like(f"%{query}%"),
                    self.Analysis.status.like(f"%{query}%"),
                )
            )
        if status:
            analysis_query = analysis_query.filter_by(status=status)
        if isinstance(deleted, bool):
            analysis_query = analysis_query.filter_by(is_deleted=deleted)
        if temp:
            analysis_query = analysis_query.filter(self.Analysis.status.in_(ONGOING_STATUSES))
        if before:
            analysis_query = analysis_query.filter(self.Analysis.started_at < before)
        if is_visible is not None:
            analysis_query = analysis_query.filter_by(is_visible=is_visible)
        return analysis_query.order_by(self.Analysis.started_at.desc())

    def analysis(self, analysis_id: int) -> models.Analysis:
        """Get a single analysis."""
        return self.Analysis.query.get(analysis_id)

    def track_update(self):
        """Update the latest updated date in the database."""
        metadata = self.info()
        metadata.updated_at = dt.datetime.now()
        self.commit()

    def is_latest_analysis_ongoing(self, case_id: str) -> bool:
        """Check if the latest analysis is ongoing for a case_id"""
        latest_analysis = self.analyses(family=case_id).first()
        if latest_analysis and latest_analysis.status in ONGOING_STATUSES:
            return True
        return False

    def is_latest_analysis_failed(self, case_id: str) -> bool:
        """Check if the latest analysis is failed for a case_id"""
        latest_analysis = self.analyses(family=case_id).first()
        if latest_analysis and latest_analysis.status == FAILED_STATUS:
            return True
        return False

    def get_latest_analysis_status(self, case_id: str) -> str:
        """Get latest analysis status for a case_id"""
        latest_analysis = self.analyses(family=case_id).first()
        return latest_analysis.status

    def is_latest_analysis_completed(self, case_id: str) -> bool:
        """Check if the latest analysis is completed for a case_id"""
        latest_analysis = self.analyses(family=case_id).first()
        if latest_analysis and latest_analysis.status == COMPLETED_STATUS:
            return True
        return False

    def info(self) -> models.Info:
        """Return metadata entry."""
        return self.Info.query.first()

    def add_pending(self, family: str, email: str = None) -> models.Analysis:
        """Add pending entry for an analysis."""
        started_at = dt.datetime.now()
        new_log = self.Analysis(family=family, status="pending", started_at=started_at)
        new_log.user = self.user(email) if email else None
        self.add_commit(new_log)
        return new_log

    def add_user(self, name: str, email: str) -> models.User:
        """Add a new user to the database."""
        new_user = self.User(name=name, email=email)
        self.add_commit(new_user)
        return new_user

    def user(self, email: str) -> models.User:
        """Fetch a user from the database."""
        return self.User.query.filter_by(email=email).first()

    def aggregate_failed(self, since_when: dt.date = None) -> List:
        """Count the number of failed jobs per category (name)."""

        categories = self.session.query(
            self.Job.name.label("name"),
            sqa.func.count(self.Job.id).label("count"),
        ).filter(self.Job.status != "cancelled")

        if since_when:
            categories = categories.filter(self.Job.started_at > since_when)

        categories = categories.group_by(self.Job.name).all()

        data = [{"name": category.name, "count": category.count} for category in categories]
        return data

    def jobs(self):
        """Return all jobs in the database."""
        return self.Job.query

    # CG code

    def mark_analyses_deleted(self, case_id: str) -> None:
        """ mark analyses connected to a case as deleted """
        for old_analysis in self.analyses(family=case_id):
            old_analysis.is_deleted = True
        self.commit()

    # Duplicate/redundant needed for backwards compatibility for now
    def add_pending_analysis(self, case_id: str, email: str = None) -> None:
        """ Add analysis as pending"""
        self.add_pending(case_id, email)

    @staticmethod
    def get_sampleinfo(analysis: models.Analysis) -> str:
        """Get the sample info path for an analysis."""
        raw_data = ruamel.yaml.safe_load(Path(analysis.config_path).open())
        data = files.parse_config(raw_data)
        return data["sampleinfo_path"]

    @staticmethod
    def parse_qcmetrics(data: dict) -> dict:
        """Call internal Trailblazer MIP API."""
        return files.parse_qcmetrics(data)

    # Duplicate/redundant needed for backwards compatibility for now
    def is_analysis_ongoing(self, case_id: str) -> bool:
        """Call internal Trailblazer API"""
        return self.is_latest_analysis_ongoing(case_id=case_id)

    # Duplicate
    def is_analysis_failed(self, case_id: str) -> bool:
        """Call internal Trailblazer API"""
        return self.is_latest_analysis_failed(case_id=case_id)

    # Duplicate/redundant needed for backwards compatibility for now
    def is_analysis_completed(self, case_id: str) -> bool:
        """Call internal Trailblazer API"""
        return self.is_latest_analysis_completed(case_id=case_id)

    # Duplicate/redundant needed for backwards compatibility for now
    def get_analysis_status(self, case_id: str) -> str:
        """Call internal Trailblazer API"""
        return self.get_latest_analysis_status(case_id=case_id)

    def has_analysis_started(self, case_id: str) -> bool:
        """Check if analysis has started"""
        statuses = ("ongoing", "failed", "completed")
        get_analysis_status = {
            "ongoing": self.is_analysis_ongoing,
            "failed": self.is_analysis_failed,
            "completed": self.is_analysis_completed,
        }
        for status in statuses:
            has_started = get_analysis_status[status](case_id=case_id)
            if has_started:
                return has_started
        return False

    def delete_analysis(
        self,
        family: str,
        date: dt.datetime,
        yes: bool = False,
        dry_run: bool = False,
    ):
        """Delete the analysis output."""
        if self.analyses(family=family, temp=True).count() > 0:
            raise ValueError("analysis for family already running")
        analysis_obj = self.find_analysis(family, date, "completed")
        assert analysis_obj.is_deleted is False
        analysis_path = Path(analysis_obj.out_dir).parent

        if yes and not dry_run:
            shutil.rmtree(analysis_path, ignore_errors=True)
            analysis_obj.is_deleted = True
            self.commit()

    @staticmethod
    def get_trending(mip_config_raw: str, qcmetrics_raw: str, sampleinfo_raw: dict) -> dict:
        """Get trending data for a MIP analysis"""
        return trending.parse_mip_analysis(
            mip_config_raw=mip_config_raw,
            qcmetrics_raw=qcmetrics_raw,
            sampleinfo_raw=sampleinfo_raw,
        )

    def get_family_root_dir(self, family_id: str):
        """Get path for a case"""
        return Path(self.families_dir) / family_id

    def get_latest_logged_analysis(self, case_id: str):
        """Get the the analysis with the latest logged_at date"""
        return self.analyses(family=case_id).order_by(models.Analysis.logged_at.desc())

    @staticmethod
    def get_sampleinfo_date(data: dict) -> str:
        """Get date from a sampleinfo """
        return data["analysis_date"]


class Store(alchy.Manager, BaseHandler):
    def __init__(self, uri: str, families_dir: str):
        super(Store, self).__init__(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=models.Model)
        self.families_dir = families_dir



# -*- coding: utf-8 -*-
import datetime

from flask import abort, g, Blueprint, jsonify, make_response, request
from google.auth import jwt

from trailblazer.server.ext import store

blueprint = Blueprint("api", __name__, url_prefix="/api/v1")


@blueprint.before_request
def before_request():
    if request.method == "OPTIONS":
        return make_response(jsonify(ok=True), 204)
    auth_header = request.headers.get("Authorization")
    if auth_header:
        jwt_token = auth_header.split("Bearer ")[-1]
    else:
        return abort(403, "no JWT token found on request")
    user_data = jwt.decode(jwt_token, verify=False)
    user_obj = store.user(user_data["email"])
    if user_obj is None:
        return abort(403, f"{user_data['email']} doesn't have access")
    g.current_user = user_obj


@blueprint.route("/analyses")
def analyses():
    """Display analyses."""
    per_page = int(request.args.get("per_page", 50))
    page = int(request.args.get("page", 1))
    query = store.analyses(
        status=request.args.get("status"),
        query=request.args.get("query"),
        is_visible=request.args.get("is_visible") == "true" or None,
    )

    query_page = query.paginate(page, per_page=per_page)
    data = []
    for analysis_obj in query_page.items:
        analysis_data = analysis_obj.to_dict()
        analysis_data["user"] = analysis_obj.user.to_dict() if analysis_obj.user else None
        analysis_data["failed_jobs"] = [job_obj.to_dict() for job_obj in analysis_obj.failed_jobs]
        data.append(analysis_data)

    return jsonify(analyses=data)


@blueprint.route("/analyses/<int:analysis_id>", methods=["GET", "PUT"])
def analysis(analysis_id):
    """Display a single analysis."""
    analysis_obj = store.analysis(analysis_id)
    if analysis_obj is None:
        return abort(404)

    if request.method == "PUT":
        analysis_obj.update(request.json)
        store.commit()

    data = analysis_obj.to_dict()
    data["failed_jobs"] = [job_obj.to_dict() for job_obj in analysis_obj.failed_jobs]
    data["user"] = analysis_obj.user.to_dict() if analysis_obj.user else None
    return jsonify(**data)


@blueprint.route("/info")
def info():
    """Display meta data about database."""
    metadata_obj = store.info()
    return jsonify(**metadata_obj.to_dict())


@blueprint.route("/me")
def me():
    """Return information about a logged in user."""
    return jsonify(**g.current_user.to_dict())


@blueprint.route("/aggregate/jobs")
def aggregate_jobs():
    """Return stats about jobs."""
    days_back = int(request.args.get("days_back", 31))
    one_month_ago = datetime.datetime.now() - datetime.timedelta(days=days_back)

    data = store.aggregate_failed(one_month_ago)
    return jsonify(jobs=data)



"""MIP specific fixtures"""
import pytest

from trailblazer.mip import sacct


@pytest.fixture(scope="session")
def failed_sacct_jobs():
    """Get failed jobs ids"""
    with open("tests/fixtures/sacct/failed.log.status") as stream:
        sacct_jobs = sacct.parse_sacct(stream)
    return sacct_jobs


@pytest.fixture(scope="session")
def mip_meta_data() -> dict:
    """Define MIP meta data metrics"""
    return {
        "FATHER_AT_DROPOUT": 2.673848,
        "FATHER_FRACTION_DUPLICATES": 0.0400685961424888,
        "FATHER_GC_DROPOUT": 0.198037,
        "FATHER_MEDIAN_INSERT_SIZE": 393,
        "FATHER_PCT_TARGET_BASES_10X": 0.987132,
        "FATHER_PCT_TARGET_BASES_20X": 0.916531,
        "FATHER_PCT_TARGET_BASES_50X": 0.004152,
        "FATHER_PCT_TARGET_BASES_100X": 0.000118,
        "FATHER_MEAN_TARGET_COVERAGE": 29.027266,
        "FATHER_STANDARD_DEVIATION": 88.653614,
        "FATHER_STRAND_BALANCE": 0.501377,
        "GENOME_BUILD_SOURCE": "grch",
        "GENOME_BUILD_VERSION": 37,
        "MOTHER_AT_DROPOUT": 1.716704,
        "MOTHER_FRACTION_DUPLICATES": 0.0379523291229131,
        "MOTHER_GC_DROPOUT": 0.214813,
        "MOTHER_MEDIAN_INSERT_SIZE": 409,
        "MOTHER_PCT_TARGET_BASES_10X": 0.98974,
        "MOTHER_PCT_TARGET_BASES_20X": 0.935455,
        "MOTHER_PCT_TARGET_BASES_50X": 0.002685,
        "MOTHER_PCT_TARGET_BASES_100X": 0.000101,
        "MOTHER_MEAN_TARGET_COVERAGE": 28.643247,
        "MOTHER_RAW_TOTAL_SEQUENCES": 600006004,
        "MOTHER_READS_MAPPED": 598456583,
        "MOTHER_STANDARD_DEVIATION": 94.353778,
        "MOTHER_STRAND_BALANCE": 0.50162,
        "MOTHER_MAPPED": 598456583 / 600006004,
        "RANK_MODEL_VERSION": "1.25",
    }



"""Parse the MIP config, qc_metric and qc_sampleinfo file"""

from typing import List, Optional


def get_case_from_config(config: dict) -> Optional[str]:
    """Get case id from config"""
    if "case_id" in config:
        return config["case_id"]
    if "family_id" in config:  # family for MIP<7
        return config["family_id"]
    return None


def get_dry_run_all(config: dict) -> bool:
    """Get dry_run_all from config"""
    if "dry_run_all" in config:
        return True
    return False


def get_sample_data_from_config(config: dict) -> List[dict]:
    """Get sample data from config"""
    samples = []
    for sample_id, analysis_type in config["analysis_type"].items():
        samples.append({"id": sample_id, "type": analysis_type})
    return samples


def parse_config(data: dict) -> dict:
    """Parse MIP config file

    Args:
        data (dict): raw YAML input from MIP analysis config file

    Returns:
        dict: parsed data
    """
    return {
        "email": data.get("email"),
        "case": get_case_from_config(config=data),
        "samples": get_sample_data_from_config(config=data),
        "config_path": data["config_file_analysis"],
        "is_dryrun": get_dry_run_all(config=data),
        "log_path": data["log_file"],
        "out_dir": data["outdata_dir"],
        "priority": data["slurm_quality_of_service"],
        "sampleinfo_path": data["sample_info_file"],
    }


def get_analysisrunstatus(sample_info: dict) -> bool:
    """Get analysis run status from sample info"""
    if sample_info["analysisrunstatus"] == "finished":
        return True
    return False


def get_sampleinfo_date(data: dict) -> str:
    """Get MIP sample info date

    Args:
        data (dict): raw YAML input from MIP qc sample info file

    Returns:
        str: analysis date
    """

    return data["analysis_date"]


def parse_sampleinfo_light(data: dict) -> dict:
    """Parse MIP sample info file and retrieve mip_version, date, and status

    Args:
        data (dict): raw YAML input from MIP qc sample info file

    Returns:
        dict: {'version': str, 'date': str, 'is_finished': str}


    """

    outdata = {
        "date": get_sampleinfo_date(data=data),
        "version": data["mip_version"],
        "is_finished": get_analysisrunstatus(sample_info=data),
    }

    return outdata


def get_case_from_sampleinfo(sample_info: dict) -> str:
    """Get case id from sampleinfo"""
    if "case" in sample_info:
        return sample_info["case"]
    if "family" in sample_info:  # family for MIP<7
        return sample_info["family"]
    return None


def get_rank_model_version(sample_info: dict, rank_model_type: str, step: str) -> str:
    """Get rank model version"""
    rank_model_version = None
    for key in ("recipe", "program"):
        if key in sample_info:
            rank_model_version = sample_info[key][step][rank_model_type]["version"]
            break
    return rank_model_version


def get_genome_build(sample_info: dict) -> str:
    """Get genome build"""
    genome_build = sample_info["human_genome_build"]
    return f"{genome_build['source']}{genome_build['version']}"


def parse_sampleinfo(data: dict) -> dict:
    """Parse MIP sample info file.

    Args:
        data (dict): raw YAML input from MIP qc sample info file

    Returns:
        dict: parsed data
    """
    outdata = {
        "date": get_sampleinfo_date(data=data),
        "genome_build": get_genome_build(sample_info=data),
        "case": get_case_from_sampleinfo(sample_info=data),
        "is_finished": get_analysisrunstatus(sample_info=data),
        "rank_model_version": get_rank_model_version(
            sample_info=data, rank_model_type="rank_model", step="genmod"
        ),
        "samples": [],
        "sv_rank_model_version": get_rank_model_version(
            sample_info=data, rank_model_type="sv_rank_model", step="sv_genmod"
        ),
        "version": data["mip_version"],
    }

    for sample_id in data["sample"].items():
        sample = {"id": sample_id}
        outdata["samples"].append(sample)

    return outdata


def set_bamstats_metrics(file_metrics: dict, sample_data: dict) -> dict:
    """Set bamstats metrics"""
    total_reads = sample_data["reads"] if "reads" in sample_data else 0
    sample_data["reads"] = int(file_metrics["bamstats"]["raw_total_sequences"]) + total_reads

    total_mapped = sample_data["total_mapped"] if "total_mapped" in sample_data else 0
    sample_data["total_mapped"] = int(file_metrics["bamstats"]["reads_mapped"]) + total_mapped
    return sample_data


def set_chanjo_sexcheck_metrics(file_metrics: dict, sample_data: dict) -> dict:
    """Set chanjo_sexcheck metrics"""
    sample_data["predicted_sex"] = file_metrics["chanjo_sexcheck"]["gender"]
    return sample_data


def set_collecthsmetrics_metrics(file_metrics: dict, sample_data: dict) -> dict:
    """Set collecthsmetrics metrics"""
    hs_metrics = file_metrics["collecthsmetrics"]["header"]["data"]
    sample_data["at_dropout"] = float(hs_metrics["AT_DROPOUT"])
    sample_data["completeness_target"] = {
        10: float(hs_metrics["PCT_TARGET_BASES_10X"]),
        20: float(hs_metrics["PCT_TARGET_BASES_20X"]),
        50: float(hs_metrics["PCT_TARGET_BASES_50X"]),
        100: float(hs_metrics["PCT_TARGET_BASES_100X"]),
    }
    sample_data["gc_dropout"] = float(hs_metrics["GC_DROPOUT"])
    sample_data["target_coverage"] = float(hs_metrics["MEAN_TARGET_COVERAGE"])
    return sample_data


def set_collectmultiplemetrics_metrics(file_metrics: dict, sample_data: dict) -> dict:
    """Set collectmultiplemetrics metrics"""
    mm_metrics = file_metrics["collectmultiplemetrics"]["header"]["pair"]
    sample_data["strand_balance"] = float(mm_metrics["STRAND_BALANCE"])
    return sample_data


def set_collectmultiplemetricsinsertsize_metrics(file_metrics: dict, sample_data: dict) -> dict:
    """Set collectmultiplemetricsinsertsize metrics"""
    mm_insert_metrics = file_metrics["collectmultiplemetricsinsertsize"]["header"]["data"]
    sample_data["median_insert_size"] = int(mm_insert_metrics["MEDIAN_INSERT_SIZE"])
    sample_data["insert_size_standard_deviation"] = float(mm_insert_metrics["STANDARD_DEVIATION"])
    return sample_data


def set_markduplicates_metrics(file_metrics: dict, sample_data: dict) -> dict:
    """Set markduplicates metrics"""
    sample_data["duplicates"] = float(file_metrics["markduplicates"]["fraction_duplicates"])
    return sample_data


def get_sample_metrics(sample_metrics: dict, sample_data: dict) -> dict:
    """Get tool qc metrics from sample metrics"""
    get_metrics = {
        "bamstats": set_bamstats_metrics,
        "chanjo_sexcheck": set_chanjo_sexcheck_metrics,
        "collecthsmetrics": set_collecthsmetrics_metrics,
        "collectmultiplemetrics": set_collectmultiplemetrics_metrics,
        "collectmultiplemetricsinsertsize": set_collectmultiplemetricsinsertsize_metrics,
        "markduplicates": set_markduplicates_metrics,
    }

    for file_metrics in sample_metrics.values():

        for tool in file_metrics:

            if get_metrics.get(tool):
                get_metrics[tool](file_metrics=file_metrics, sample_data=sample_data)
    return sample_data


def parse_qcmetrics(metrics: dict) -> dict:
    """Parse MIP qc metrics file
    Args:
        metrics (dict): raw YAML input from MIP qc metrics file
    Returns:
        dict: parsed qc metrics metrics
    """
    qc_metric = {"samples": []}

    for sample_id, sample_metrics in metrics["sample"].items():

        sample_data = {
            "id": sample_id,
        }
        sample_data = get_sample_metrics(sample_metrics=sample_metrics, sample_data=sample_data)
        sample_data["mapped"] = sample_data["total_mapped"] / sample_data["reads"]
        qc_metric["samples"].append(sample_data)
    return qc_metric



from pathlib import Path
import shutil

import click


@click.command()
@click.option("-f", "--force", is_flag=True, help="skip sanity checks")
@click.option("-y", "--yes", is_flag=True, help="skip manual confirmations")
@click.argument("analysis_id", type=int)
@click.pass_context
def delete(context, force, yes, analysis_id):
    """Mark analysis as deleted in db and remove analysis folder on disk."""
    analysis_obj = context.obj["store"].analysis(analysis_id)
    if analysis_obj is None:
        print(click.style("analysis log not found", fg="red"))
        context.abort()

    print(click.style(f"{analysis_obj.family}: {analysis_obj.status}"))

    if analysis_obj.has_ongoing_status:
        if yes or click.confirm(f"remove analysis log?"):
            analysis_obj.delete()
            context.obj["store"].commit()
            print(click.style(f"analysis deleted: {analysis_obj.family}", fg="blue"))
    else:
        if analysis_obj.is_deleted:
            print(click.style(f"{analysis_obj.family}: already deleted", fg="red"))
            context.abort()

        if Path(analysis_obj.out_dir).exists() or force:
            root_dir = context.obj["store"].families_dir
            family_dir = analysis_obj.out_dir
            if not force and (len(family_dir) <= len(root_dir) or root_dir not in family_dir):
                print(click.style(f"unknown analysis output dir: {analysis_obj.out_dir}", fg="red"))
                print(click.style("use '--force' to override"))
                context.abort()

            if yes or click.confirm(f"remove analysis output: {analysis_obj.out_dir}?"):
                shutil.rmtree(analysis_obj.out_dir, ignore_errors=True)
                analysis_obj.is_deleted = True
                context.obj["store"].commit()
                print(click.style(f"analysis deleted: {analysis_obj.family}", fg="blue"))
        else:
            print(click.style(f"analysis output doesn't exist: {analysis_obj.out_dir}", fg="red"))
            print(click.style("use '--force' to override"))
            context.abort()



# -*- coding: utf-8 -*-
from datetime import datetime

NORMAL_CATEGORIES = ("COMPLETED", "RUNNING", "PENDING")
FAILED_CATEGORIES = ("FAILED", "CANCELLED", "TIMEOUT")
CATEGORIES = NORMAL_CATEGORIES + FAILED_CATEGORIES


def time_to_sec(time_str: str) -> int:
    """Convert time in string format to seconds.

    Skipping seconds since sometimes the last column is truncated
    for entries where >10 days.
    """
    total_sec = 0
    if "-" in time_str:
        # parse out days
        days, time_str = time_str.split("-")
        total_sec += int(days) * 24 * 60 * 60

    # parse out the hours and mins (skip seconds)
    hours_min_raw = time_str.split(":")[:-1]
    time_parts = [int(round(float(val))) for val in hours_min_raw]
    total_sec += time_parts[-1] * 60  # minutes
    if len(time_parts) > 1:
        total_sec += time_parts[-2] * 60 * 60  # hours
    return total_sec


def convert_job(row: list) -> dict:
    """Convert sacct row to dict."""
    state = row[-2]
    start_time_raw = row[-4]
    end_time_raw = row[-3]
    if state not in ("PENDING", "CANCELLED"):
        start_time = datetime.strptime(start_time_raw, "%Y-%m-%dT%H:%M:%S")
        if state != "RUNNING":
            end_time = datetime.strptime(end_time_raw, "%Y-%m-%dT%H:%M:%S")
        else:
            end_time = None
    else:
        start_time = end_time = None

    # parse name of job
    job_name = row[1]
    step_name, step_context = job_name.rstrip("_BOTH").rstrip("_SV").rsplit("_", 1)

    return {
        "id": int(row[0]),
        "name": job_name,
        "step": step_name,
        "context": step_context,
        "state": state,
        "start": start_time,
        "end": end_time,
        "elapsed": time_to_sec(row[-5]),
        "cpu": time_to_sec(row[-6]),
        "is_completed": state == "COMPLETED",
    }


def parse_sacct(sacct_stream):
    """Parse out information from sacct status output."""
    rows = (line.split() for line in sacct_stream)
    # filter rows that begin with a SLURM job id
    relevant_rows = (row for row in rows if row[0].isdigit())
    jobs = [convert_job(row) for row in relevant_rows]
    return jobs


def filter_jobs(sacct_jobs, failed=True):
    """Filter jobs that have a FAILED etc. status."""
    categories = FAILED_CATEGORIES if failed else NORMAL_CATEGORIES
    filtered_jobs = [job for job in sacct_jobs if job["state"] in categories]
    return filtered_jobs



import datetime as dt
import logging

import click
from .delete import delete

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-y", "--yes", is_flag=True, help="skip manual confirmations")
@click.option("-d", "--days-ago", default=14, help="days ago analyses were started")
@click.option("-f", "--force", is_flag=True, help="skip sanity checks")
@click.pass_context
def clean(context, days_ago, yes, force):
    """Clean up files from "old" analyses runs."""
    number_of_days_ago = dt.datetime.now() - dt.timedelta(days=days_ago)
    analyses = context.obj["store"].analyses(
        status="completed",
        before=number_of_days_ago,
        deleted=False,
    )
    for analysis_obj in analyses:
        LOG.debug(f"checking analysis: {analysis_obj.family} ({analysis_obj.id})")
        latest_analysis = context.obj["store"].analyses(family=analysis_obj.family).first()
        if analysis_obj != latest_analysis:
            print(click.style(f"{analysis_obj.family}: family has been re-started", fg="yellow"))
        else:
            print(f"delete analysis: {analysis_obj.family} ({analysis_obj.id})")
            context.invoke(delete, analysis_id=analysis_obj.id, yes=yes, force=force)



import click
from dateutil.parser import parse as parse_date

from trailblazer.store.models import STATUS_OPTIONS


@click.command("ls")
@click.option("-s", "--status", type=click.Choice(STATUS_OPTIONS))
@click.option("-b", "--before", help="return analyses started before date")
@click.pass_context
def ls_cmd(context, before, status):
    """Display recent logs for analyses."""
    runs = (
        context.obj["store"]
        .analyses(
            status=status,
            deleted=False,
            before=parse_date(before) if before else None,
        )
        .limit(30)
    )
    for run_obj in runs:
        if run_obj.status == "pending":
            message = f"{run_obj.id} | {run_obj.family} [{run_obj.status.upper()}]"
        else:
            message = (
                f"{run_obj.id} | {run_obj.family} {run_obj.started_at.date()} "
                f"[{run_obj.type.upper()}/{run_obj.status.upper()}]"
            )
            if run_obj.status == "running":
                message = click.style(f"{message} - {run_obj.progress * 100}/100", fg="blue")
            elif run_obj.status == "completed":
                message = click.style(f"{message} - {run_obj.completed_at}", fg="green")
            elif run_obj.status == "failed":
                message = click.style(message, fg="red")
        print(message)



# -*- coding: utf-8 -*-
"""Test MIP files"""

from trailblazer.mip import trending


def test_parse_mip_analysis(files_raw) -> dict:
    """
    Args:
    files_raw (dict): With dicts from files
    """
    # GIVEN config data of a "sharp" run (not dry run)
    mip_config_raw = files_raw["config"]

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = files_raw["sampleinfo"]

    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    trend_data = trending.parse_mip_analysis(mip_config_raw, qcmetrics_raw, sampleinfo_raw)

    # Build dict for family return data
    trend_test_data = {
        "case": "case",
        "genome_build": "grch37",
        "mip_version": "v7.1.0",
        "rank_model_version": "1.25",
    }

    # Check returns from def
    for key, value in trend_test_data.items():
        assert trend_data[key] == value

    # Check sample id
    assert "mother" in trend_data["sample_ids"]

    # Build dict for sample return data
    trend_test_sample_data = {
        "at_dropout": {
            "mother": 1.716704,
        },
        "analysis_sex": {
            "mother": "female",
        },
        "duplicates": {
            "mother": 3.7952329122913104,
        },
        "insert_size_standard_deviation": {
            "mother": 94.353778,
        },
        "mapped_reads": {
            "mother": 99.74176575073072,
        },
        "median_insert_size": {
            "mother": 409,
        },
        "gc_dropout": {
            "mother": 0.214813,
        },
    }

    # Check returns from def
    for metric in trend_test_sample_data:
        for sample, value in trend_test_sample_data[metric].items():
            assert trend_data[metric][sample] == value
